import pytest

from brilliance_admin.auth import AuthData
from brilliance_admin.exceptions import AdminAPIException
from brilliance_admin import sqlalchemy
from brilliance_admin.schema.admin_schema import AdminSchemaData
from example.sections.models import User, UserFactory


@pytest.mark.asyncio
async def test_login(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test', is_admin=True)
    result = await auth.login(data=AuthData(username='123', password='test'), debug=True)
    assert result.user.username == user.username


@pytest.mark.asyncio
async def test_login_not_admin(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    await UserFactory(username='123', password='test')
    with pytest.raises(AdminAPIException) as e:
        await auth.login(data=AuthData(username='123', password='test'), debug=True)

    assert e.value.get_error().code == 'not_an_admin'


@pytest.mark.asyncio
async def test_login_not_found(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    with pytest.raises(AdminAPIException) as e:
        await auth.login(data=AuthData(username='123', password='test'), debug=True)

    assert e.value.get_error().code == 'user_not_found'


@pytest.mark.asyncio
async def test_authenticate(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test', is_admin=True)

    token = auth.get_token(user)
    result_user = await auth.authenticate(headers={'Authorization': f'Token {token}'})
    AdminSchemaData(categories={}, profile=result_user)

    assert result_user.username == user.username


@pytest.mark.asyncio
async def test_authenticate_bad_secret(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test', is_admin=True)

    token = auth.get_token(user)
    result_user = await auth.authenticate(headers={'Authorization': f'Token {token}'})
    AdminSchemaData(categories={}, profile=result_user)
    assert result_user.username == user.username

    auth.secret = 'another'
    with pytest.raises(AdminAPIException) as e:
        await auth.authenticate(headers={'Authorization': f'Token {token}'})

    assert e.value.get_error().code == 'token_error'


@pytest.mark.asyncio
async def test_authenticate_not_admin(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test')
    with pytest.raises(AdminAPIException) as e:
        token = auth.get_token(user)
        await auth.authenticate(headers={'Authorization': f'Token {token}'})

    assert e.value.get_error().code == 'user_not_found'


@pytest.mark.asyncio
async def test_login_valid_password_sync(sqlite_sessionmaker, language_context):
    def async_password_validator(user, password):
        return password == 'correct_password'

    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
        password_validator=async_password_validator,
    )
    await UserFactory(username='admin', password='correct_password', is_admin=True)

    result = await auth.login(data=AuthData(username='admin', password='correct_password'))

    assert result.token is not None
    assert result.user.username == 'admin'


@pytest.mark.asyncio
async def test_login_valid_password_async(sqlite_sessionmaker, language_context):
    async def async_password_validator(user, password):
        return password == 'correct_password'

    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
        password_validator=async_password_validator,
    )
    await UserFactory(username='admin', password='correct_password', is_admin=True)

    result = await auth.login(data=AuthData(username='admin', password='correct_password'))

    assert result.token is not None
    assert result.user.username == 'admin'


@pytest.mark.asyncio
async def test_login_invalid_password_async(sqlite_sessionmaker, language_context):
    async def async_password_validator(user, password):
        return password == 'correct_password'

    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
        password_validator=async_password_validator,
    )
    await UserFactory(username='admin', password='correct_password', is_admin=True)

    with pytest.raises(AdminAPIException) as e:
        await auth.login(data=AuthData(username='admin', password='wrong_password'), debug=False)

    context = {'language_context': language_context}
    assert e.value.get_error().model_dump(context=context) == {
        'code': 'user_not_found',
        'field_errors': None,
        'message': None,
    }


@pytest.mark.asyncio
async def test_login_invalid_password(sqlite_sessionmaker, language_context):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
        password_validator=lambda user, password: password == 'correct_password',
    )
    await UserFactory(username='admin', password='correct_password', is_admin=True)

    with pytest.raises(AdminAPIException) as e:
        await auth.login(data=AuthData(username='admin', password='wrong_password'), debug=True)

    context = {'language_context': language_context}
    assert e.value.get_error().model_dump(context=context) == {
        'code': 'user_not_found',
        'field_errors': None,
        'message': None,
    }



from unittest import mock

import pytest
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from brilliance_admin import auth, schema, sqlalchemy
from brilliance_admin.exceptions import AdminAPIException
from brilliance_admin.schema import admin_schema
from brilliance_admin.translations import TranslateText
from example.sections.models import (
    Currency, CurrencyFactory, Merchant, MerchantFactory, Terminal, TerminalFactory, TerminalStatuses)
from tests.test_sqlalcmeny_schema import FIELDS


def get_category(sqlite_sessionmaker):
    category = sqlalchemy.SQLAlchemyAdmin(
        model=Terminal,
        db_async_session=sqlite_sessionmaker,
        table_schema=sqlalchemy.SQLAlchemyFieldsSchema(
            model=Terminal,
            fields=FIELDS,
        ),
    )
    return category


@pytest.mark.asyncio
async def test_create(sqlite_sessionmaker, language_context):
    category = get_category(sqlite_sessionmaker)
    user = auth.UserABC(username="test")
    merchant = await MerchantFactory()
    currency = await CurrencyFactory()

    create_data = {
        'manager_id': 0,
        'merchant_id': merchant.id,
        'currency_id': currency.id,
        'status': {'value': 'error', 'title': 'Error'},
        'description': 'test',
        'title': 'test',
        'created_at': '2026-01-20T20:33:40.055184Z',
    }
    create_result: schema.CreateResult = await category.create(
        data=create_data,
        user=user,
        language_context=language_context,
        admin_schema=admin_schema,
    )

    assert create_result.pk == 1


@pytest.mark.asyncio
async def test_create_bad_fk(sqlite_sessionmaker, language_context):
    category = get_category(sqlite_sessionmaker)
    user = auth.UserABC(username="test")
    merchant = await MerchantFactory()
    currency = await CurrencyFactory()

    create_data = {
        'manager_id': 1,
        'merchant_id': 100,
        'currency_id': currency.id,
        'status': {'value': 'error', 'title': 'Error'},
        'description': 'test',
        'title': 'test',
    }
    with pytest.raises(AdminAPIException) as e:
        schema.CreateResult = await category.create(
            data=create_data,
            user=user,
            language_context=language_context,
            admin_schema=admin_schema,
        )
    assert e.value.get_error().model_dump() == {'code': 'db_integrity_error', 'field_errors': None, 'message': 'NOT NULL constraint failed: terminal.merchant_id'}


@pytest.mark.asyncio
async def test_retrieve(sqlite_sessionmaker, language_context):
    category = get_category(sqlite_sessionmaker)
    user = auth.UserABC(username="test")
    merchant = await MerchantFactory()
    currency = await CurrencyFactory()
    terminal = await TerminalFactory(
        title="test",
        description='test',
        status=TerminalStatuses.PROCESS.value,
        is_h2h=False,
        registered_delay=None,
        merchant=merchant,
        currency=currency,
    )

    retrieve_result = await category.retrieve(
        pk=terminal.id,
        user=user,
        language_context=language_context,
        admin_schema=admin_schema,
    )
    expected_data = {
        'manager_id': mock.ANY,
        'created_at': mock.ANY,
        'description': 'test',
        'currency_id': {
            'key': currency.id,
            'title': mock.ANY,
        },
        'fee_id': None,
        'status': {
            'title': TranslateText('statuses.process'),
            'value': 'process',
        },
        'title': 'test',
        'id': terminal.id,
        'is_h2h': False,
        'merchant_id': {'key': merchant.id, 'title': mock.ANY},
        'registered_delay': None,
        'secret_key': mock.ANY,
    }
    assert retrieve_result.data == expected_data


@pytest.mark.asyncio
async def test_retrieve_currency(sqlite_sessionmaker, language_context):
    category = sqlalchemy.SQLAlchemyAdmin(
        model=Currency,
        db_async_session=sqlite_sessionmaker,
        table_schema=sqlalchemy.SQLAlchemyFieldsSchema(
            model=Currency,
            fields=[
                'id',
                'terminals',
            ],
        ),
    )
    terminals_field = category.table_schema.get_field('terminals')
    assert terminals_field._type == "related"
    assert terminals_field.rel_name == "terminals"
    assert terminals_field.many is True

    user = auth.UserABC(username="test")
    merchant = await MerchantFactory()
    currency = await CurrencyFactory()
    terminal_1 = await TerminalFactory(
        merchant=merchant,
        currency=currency,
        title='First',
    )
    terminal_2 = await TerminalFactory(
        merchant=merchant,
        currency=currency,
        title='Second',
    )

    retrieve_result = await category.retrieve(
        pk=currency.id,
        user=user,
        language_context=language_context,
        admin_schema=admin_schema,
    )
    expected_data = {
        'id': currency.id,
        'terminals': [
            {'key': terminal_1.id, 'title': 'First'},
            {'key': terminal_2.id, 'title': 'Second'},
        ],
    }
    assert retrieve_result.data == expected_data, retrieve_result.data


@pytest.mark.asyncio
async def test_create_bad_json(sqlite_sessionmaker, language_context):
    category = sqlalchemy.SQLAlchemyAdmin(
        model=Merchant,
        db_async_session=sqlite_sessionmaker,
        table_schema=sqlalchemy.SQLAlchemyFieldsSchema(
            model=Merchant,
            fields=[
                'title',
                'provider_settings',
            ],
        ),
    )
    user = auth.UserABC(username="test")
    create_data = {
        'title': 'test',
        'provider_settings': 'not json',
    }
    with pytest.raises(AdminAPIException) as e:
        await category.create(
            data=create_data,
            user=user,
            language_context=language_context,
            admin_schema=admin_schema,
        )
    context = {'language_context': language_context}
    errors = {
        'code': 'validation_error',
        'field_errors': {
            'provider_settings': {
                'code': None,
                'field_slug': None,
                'message': "Некорректный тип данных: <class 'str'>; ожидается JSON",
            },
        },
        'message': 'Validation error',
    }
    assert e.value.get_error().model_dump(context=context) == errors


@pytest.mark.asyncio
async def test_list(sqlite_sessionmaker, language_context):
    category = get_category(sqlite_sessionmaker)
    user = auth.UserABC(username="test")
    await TerminalFactory(
        is_h2h=False,
        registered_delay=None,
        title='Test terminal',
        description="description",
        status=TerminalStatuses.PROCESS.value,
        merchant=await MerchantFactory(title="Test merch"),
        currency=await CurrencyFactory(),
    )

    list_result: dict = await category.get_list(
        list_data=schema.ListData(
            filters={
                'id': '',
            }
        ),
        user=user,
        language_context=language_context,
        admin_schema=admin_schema,
    )
    data = [
        {
            'manager_id': mock.ANY,
            'created_at': mock.ANY,
            'currency_id': {
                'key': 1,
                'title': mock.ANY,
            },
            'status': {
                'title': TranslateText('statuses.process'),
                'value': 'process',
            },
            'fee_id': None,
            'description': 'description',
            'id': 1,
            'is_h2h': False,
            'merchant_id': {
                'key': 1,
                'title': mock.ANY,
            },
            'registered_delay': None,
            'secret_key': mock.ANY,
            'title': 'Test terminal',
        },
    ]
    expected_create = schema.TableListResult(
        data=data,
        total_count=1,
    )
    assert list_result == expected_create


@pytest.mark.asyncio
async def test_update_related_one(sqlite_sessionmaker, language_context):
    category = get_category(sqlite_sessionmaker)
    user = auth.UserABC(username="test")
    terminal = await TerminalFactory(
        merchant=await MerchantFactory(title="Test merch"),
        currency=await CurrencyFactory(),
    )
    new_merchant = await MerchantFactory(title="New merch")

    update_data = {
        'merchant_id': {'key': new_merchant.id, 'title': '123'},
        'description': 'new description',
    }
    update_result = await category.update(
        pk=terminal.id,
        data=update_data,
        user=user,
        language_context=language_context,
        admin_schema=admin_schema,
    )
    assert update_result == schema.UpdateResult(pk=terminal.id)


@pytest.mark.asyncio
async def test_update_related_many(sqlite_sessionmaker, language_context):
    category = sqlalchemy.SQLAlchemyAdmin(
        model=Currency,
        db_async_session=sqlite_sessionmaker,
        table_schema=sqlalchemy.SQLAlchemyFieldsSchema(
            model=Currency,
            fields=[
                'id',
                'terminals',
            ],
        ),
    )
    user = auth.UserABC(username="test")

    currency_rub = await CurrencyFactory(title='RUB')
    currency_usd = await CurrencyFactory(title='USD')
    terminal_1 = await TerminalFactory(
        merchant=await MerchantFactory(title="Test merch"),
        currency=currency_usd,
    )
    terminal_2 = await TerminalFactory(
        merchant=await MerchantFactory(title="Test merch"),
        currency=currency_usd,
    )

    update_data = {
        'terminals': [
            {'key': terminal_1.id, 'title': 'test'},
            {'key': terminal_2.id, 'title': 'test'},
        ],
    }
    update_result = await category.update(
        pk=currency_rub.id,
        data=update_data,
        user=user,
        language_context=language_context,
        admin_schema=admin_schema,
    )
    assert update_result == schema.UpdateResult(pk=currency_rub.id)

    async with sqlite_sessionmaker() as session:
        updated_rub = (await session.execute(
            select(Currency)
            .options(selectinload(Currency.terminals))
            .where(Currency.id == currency_rub.id)
        )).scalar_one()

        updated_usd = (await session.execute(
            select(Currency)
            .options(selectinload(Currency.terminals))
            .where(Currency.id == currency_usd.id)
        )).scalar_one()

    assert sorted(t.id for t in updated_rub.terminals) == [terminal_1.id, terminal_2.id]
    assert sorted(t.id for t in updated_usd.terminals) == []


@pytest.mark.asyncio
async def test_autocomplete(sqlite_sessionmaker, language_context):
    category = get_category(sqlite_sessionmaker)
    category = sqlalchemy.SQLAlchemyAdmin(model=Terminal, db_async_session=sqlite_sessionmaker)

    user = auth.UserABC(username="test")
    autocomplete_result = await category.autocomplete(
        data=schema.AutocompleteData(
            field_slug='merchant_id',
        ),
        user=user,
        language_context=language_context,
        admin_schema=admin_schema,
    )
    assert autocomplete_result == schema.AutocompleteResult()



from brilliance_admin.auth import UserABC
from brilliance_admin.schema.admin_schema import AdminSchema
from brilliance_admin.schema.table.table_models import AutocompleteData, AutocompleteResult
from brilliance_admin.translations import LanguageContext


class SQLAlchemyAdminAutocompleteMixin:
    async def autocomplete(
            self,
            data: AutocompleteData,
            user: UserABC,
            language_context: LanguageContext,
            admin_schema: AdminSchema,
    ) -> AutocompleteResult:
        form_schema = None

        if data.action_name is not None:
            action_fn = self._get_action_fn(data.action_name)
            if not action_fn:
                raise Exception(f'Action "{data.action_name}" is not found')

            if not action_fn.form_schema:
                raise Exception(f'Action "{data.action_name}" form_schema is None')

            form_schema = action_fn.form_schema

        elif data.is_filter:
            if not self.table_filters:
                raise Exception(f'Action "{data.action_name}" table_filters is None')

            form_schema = self.table_filters

        else:
            form_schema = self.table_schema

        field = form_schema.get_field(data.field_slug)
        if not field:
            raise Exception(f'Field "{data.field_slug}" is not found')

        results = await field.autocomplete(
            self.model,
            data,
            user,
            extra={'db_async_session': self.db_async_session},
        )

        return AutocompleteResult(results=results)



from typing import Any

from brilliance_admin import auth, schema
from brilliance_admin.exceptions import AdminAPIException, APIError
from brilliance_admin.schema.admin_schema import AdminSchema
from brilliance_admin.translations import LanguageContext
from brilliance_admin.translations import TranslateText as _
from brilliance_admin.utils import get_logger

logger = get_logger()


class SQLAlchemyAdminUpdate:
    has_update: bool = True

    # pylint: disable=too-many-locals
    async def update(
            self,
            pk: Any,
            data: dict,
            user: auth.UserABC,
            language_context: LanguageContext,
            admin_schema: AdminSchema,
    ) -> schema.UpdateResult:
        if not self.has_update:
            raise AdminAPIException(APIError(message=_('errors.method_not_allowed')), status_code=500)

        # pylint: disable=import-outside-toplevel
        from sqlalchemy import inspect
        from sqlalchemy.exc import IntegrityError

        if pk is None:
            raise AdminAPIException(
                APIError(message=_('errors.pk_not_found') % {'pk_name': self.pk_name}, code='pk_not_found'),
                status_code=400,
            )

        col = inspect(self.table_schema.model).mapper.columns[self.pk_name]
        python_type = col.type.python_type

        stmt = self.get_queryset().where(getattr(self.model, self.pk_name) == python_type(pk))

        try:
            async with self.db_async_session() as session:
                record = (await session.execute(stmt)).scalars().first()
                if record is None:
                    msg = _('errors.record_not_found') % {'pk_name': self.pk_name, 'pk': pk}
                    raise AdminAPIException(
                        APIError(message=msg, code='record_not_found'),
                        status_code=400,
                    )

                await self.table_schema.update(record, user, data, session)

        except AdminAPIException as e:
            raise e

        except ConnectionRefusedError as e:
            logger.exception(
                'SQLAlchemy %s update %s #%s db connection error: %s',
                type(self).__name__, self.table_schema.model.__name__, pk, e,
                extra={'data': data},
            )
            msg = _('errors.connection_refused_error') % {'error': str(e)}
            raise AdminAPIException(
                APIError(message=msg, code='connection_refused_error'),
                status_code=400,
            ) from e

        except IntegrityError as e:
            logger.warning(
                'SQLAlchemy %s update %s #%s db error: %s',
                type(self).__name__, self.table_schema.model.__name__, pk, e,
                extra={'data': data},
            )
            orig = e.orig
            message = orig.args[0] if orig.args else type(orig).__name__
            raise AdminAPIException(
                APIError(message=message, code='db_integrity_error'), status_code=500,
            ) from e

        except Exception as e:
            logger.exception(
                'SQLAlchemy %s update %s #%s db error: %s',
                type(self).__name__, self.table_schema.model.__name__, pk, e,
                extra={'data': data}
            )
            msg = _('errors.db_error_update') % {
                'error_type': str(e) if admin_schema.debug else type(e).__name__,
            }
            raise AdminAPIException(
                APIError(message=msg, code='db_error_update'), status_code=500,
            ) from e

        logger.info(
            '%s model %s #%s updated by %s',
            type(self).__name__, self.table_schema.model.__name__, pk, user.username,
            extra={'data': data},
        )
        return schema.UpdateResult(pk=pk)



import abc
from importlib import resources
from typing import Any, Dict

import pydantic
from asgiref.local import Local
from pydantic_core import core_schema

from brilliance_admin.utils import DataclassBase, YamlI18n, get_logger

logger = get_logger()

_active = Local()


@pydantic.dataclasses.dataclass
class TranslateText(DataclassBase):
    slug: str
    translation_kwargs: dict | None = None

    def __init__(self, slug: str):
        self.slug = slug

    def __hash__(self):
        return hash(self.slug)

    @pydantic.model_serializer(mode='plain')
    def serialize_model(self, info: pydantic.SerializationInfo) -> str:
        ctx = info.context or {}
        language_context = ctx.get('language_context')

        if not language_context:
            raise AttributeError('language_context is not in context manager for serialization')

        if not issubclass(type(language_context), LanguageContext):
            raise AttributeError(f'language_context "{type(language_context)}" is not subclass of LanguageContext')

        return language_context.get_text(self)

    def __str__(self):
        lm = getattr(_active, '_language_context', None)
        if not lm:

            raise AttributeError(f'language_context is not in local scope for translation: {locals()}')

        if not issubclass(type(lm), LanguageContext):
            raise AttributeError(f'language_context "{lm}" is not subclass of LanguageContext')

        return lm.get_text(self)

    def __mod__(self, other):
        if not isinstance(other, dict):
            msg = f'TranslateText only dict is supported trough % operand (slug="{self.slug}" other={type(other)})'
            raise AttributeError(msg)
        self.translation_kwargs = other
        return self


class LanguageManager(abc.ABC):
    languages: Dict[str, str] | None
    phrases: YamlI18n = None

    def __init__(self, languages: str | None, locales_dir: str | None = None):
        self.languages = languages
        self.phrases = YamlI18n()

        builtin_locales_dir = resources.files("brilliance_admin").joinpath("locales")
        self.phrases.load_folder(builtin_locales_dir)
        logger.debug('Language manager builtin dir loaded: %s', builtin_locales_dir)

        if locales_dir:
            self.phrases.load_folder(locales_dir)
            logger.debug('Language manager locales_dir loaded: %s', locales_dir)

        langs = ', '.join(self.phrases.data.keys())
        logger.debug('Language manager setup completed; languages: %s', langs)

    def get_text(self, text, language) -> str:
        if not isinstance(text, TranslateText):
            return text

        default_lang = list(self.languages.keys())[0]

        translation = self.phrases.get_text(text.slug, language or default_lang, default_lang) or text.slug
        if text.translation_kwargs:
            translation %= text.translation_kwargs

        return translation

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: Any) -> core_schema.CoreSchema:
        def validate(v: Any) -> "LanguageManager":
            if isinstance(v, cls):
                return v
            raise TypeError(f"Expected {cls.__name__} instance")

        return core_schema.no_info_plain_validator_function(
            validate,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: repr(v),
                info_arg=False,
                return_schema=core_schema.str_schema(),
            ),
        )


class LanguageContext(abc.ABC):
    language: str | None
    language_manager: LanguageManager

    def __init__(self, language, language_manager):
        self.language = language
        self.language_manager = language_manager

        _active._language_context = self

    def get_text(self, text) -> str:
        return self.language_manager.get_text(text, self.language)
